import regex as re

from banks import Prompt


def test_canary_word_generation():
    p = Prompt("{{canary_word}}This is my prompt")
    assert re.match(r"BANKS\[.{8}\]This is my prompt", p.text())


def test_canary_word_leaked():
    p = Prompt("{{canary_word}}This is my prompt")
    assert p.canary_leaked(p.text())

    p = Prompt("This is my prompt")
    assert not p.canary_leaked(p.text())



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
from typing import cast

from jinja2 import nodes
from jinja2.ext import Extension
from litellm import ModelResponse, acompletion, completion

from banks.errors import CanaryWordError
from banks.prompt import Prompt

DEFAULT_MODEL = "gpt-3.5-turbo"
SYSTEM_PROMPT = Prompt("{{canary_word}} You are a helpful assistant.")


def generate(model_name: str):  # noqa
    """
    `generate` can be used to call the LiteLLM API passing the tag text as a prompt and get back some content.

    Example:
        ```
        {% generate "write a tweet with positive sentiment" "gpt-3.5-turbo" %}
        Feeling grateful for all the opportunities that come my way! #positivity #productivity
        ```
    """


class GenerateExtension(Extension):
    # a set of names that trigger the extension.
    tags = {"generate"}  # noqa

    def parse(self, parser):
        # We get the line number of the first token so that we can give
        # that line number to the nodes we create by hand.
        lineno = next(parser.stream).lineno

        # The args passed to the extension:
        # - the prompt text used to generate new text
        # - (optional) the name of the model use to generate new text
        args = [parser.parse_expression()]

        # If there is a comma, the user provided the model name. If not, use
        # None as the second parameter.
        if parser.stream.skip_if("comma"):
            args.append(parser.parse_expression())
        else:
            args.append(nodes.Const(None))

        if parser.environment.is_async:
            return nodes.Output([self.call_method("_agenerate", args)]).set_lineno(lineno)
        return nodes.Output([self.call_method("_generate", args)]).set_lineno(lineno)

    def _generate(self, text, model_name=DEFAULT_MODEL):
        """
        Helper callback.

        To tweak the prompt used to generate content, change the variable `messages` .
        """
        messages = [
            {"role": "system", "content": SYSTEM_PROMPT.text()},
            {"role": "user", "content": text},
        ]
        response: ModelResponse = cast(ModelResponse, completion(model=model_name, messages=messages))
        return self._get_content(response)

    async def _agenerate(self, text, model_name=DEFAULT_MODEL):
        """
        Helper callback.

        To tweak the prompt used to generate content, change the variable `messages` .
        """
        messages = [
            {"role": "system", "content": SYSTEM_PROMPT.text()},
            {"role": "user", "content": text},
        ]
        response: ModelResponse = cast(ModelResponse, await acompletion(model=model_name, messages=messages))
        return self._get_content(response)

    def _get_content(self, response: ModelResponse) -> str:
        content = response["choices"][0]["message"]["content"]
        if SYSTEM_PROMPT.canary_leaked(content):
            msg = "The system prompt has leaked into the response, possible prompt injection!"
            raise CanaryWordError(msg)
        return content



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
class MissingDependencyError(Exception):
    pass


class AsyncError(Exception):
    pass


class CanaryWordError(Exception):
    pass



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
from banks.env import env
from banks.prompt import AsyncPrompt, Prompt

__all__ = (
    "env",
    "Prompt",
    "AsyncPrompt",
)



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
from typing import Optional

from .config import async_enabled
from .env import env
from .errors import AsyncError
from .utils import generate_canary_word


class BasePrompt:
    def __init__(self, text: str, canary_word: Optional[str] = None) -> None:
        """
        Prompt constructor.

        Parameters:
            text: The template text
            canary_word: The string to use for the `{{canary_word}}` extension. If `None`, a default string will be
                generated

        """
        self._template = env.from_string(text)
        self.defaults = {"canary_word": canary_word or generate_canary_word()}

    def _get_context(self, data: Optional[dict]) -> dict:
        if data is None:
            return self.defaults
        return data | self.defaults

    def canary_leaked(self, text: str) -> bool:
        """
        Returns whether the canary word is present in `text`, signalling the prompt might have leaked.
        """
        return self.defaults["canary_word"] in text

    @classmethod
    def from_template(cls, name: str) -> "BasePrompt":
        """
        Create a prompt instance from a template.

        Prompt templates can be really long and at some point you might want to store them on files. To avoid the
        boilerplate code to read a file and pass the content as strings to the constructor, `Prompt`s can be
        initialized by just passing the name of the template file, provided that the file is stored in a folder called
        `templates` in the current path:

        ```
        .
        └── templates
            └── foo.jinja
        ```

        The code would be the following:

        ```py
        from banks import Prompt

        p = Prompt.from_template("foo.jinja")
        prompt_text = p.text(data={"foo": "bar"})
        ```

        !!! warning
            Banks comes with its own set of default templates (see below) which takes precedence over the
            ones loaded from the filesystem, so be sure to use different names for your custom
            templates

        Parameters:
            name: The name of the template.

        Returns:
            A new `Prompt` instance.
        """
        p = cls("")
        p._template = env.get_template(name)
        return p


class Prompt(BasePrompt):
    """
    The `Prompt` class is the only thing you need to know about Banks on the Python side. The class can be
    initialized with a string variable containing the prompt template text, then you can invoke the method `text`
    on your instance to pass the data needed to render the template and get back the final prompt.

    A quick example:
    ```py
    from banks import Prompt


    p = Prompt("Write a 500-word blog post on {{ topic }}.")
    my_topic = "retrogame computing"
    print(p.text({"topic": my_topic}))
    ```
    """

    def text(self, data: Optional[dict] = None) -> str:
        """
        Render the prompt using variables present in `data`

        Parameters:
            data: A dictionary containing the context variables.
        """
        data = self._get_context(data)
        return self._template.render(data)


class AsyncPrompt(BasePrompt):
    """
    Banks provides async support through the machinery [provided by Jinja](https://jinja.palletsprojects.com/en/3.0.x/api/#async-support)

    Since the Jinja environment is a global state in banks, the library can work either with or
    without async support, and this must be known before importing anything.

    If the application using banks runs within an `asyncio` loop, you can do two things
    to optimize banks' execution:

    1. Set the environment variable `BANKS_ASYNC_ENABLED=true`.
    2. Use the `AsyncPrompt` class that has an awaitable `run` method.

    For example, let's render a prompt that contains some calls to the `generate` extension. Those calls
    will be heavily I/O bound, so other tasks can take advantage and advance while the prompt is being
    rendered.

    Example:
    ```python
    # Enable async support before importing from banks
    import os

    os.environ["BANKS_ASYNC_ENABLED"] = "true"

    # Show logs to see what's happening at runtime
    import logging

    logging.basicConfig(level=logging.INFO)

    import asyncio
    from banks import AsyncPrompt

    prompt_template = \"\"\"
    Generate a tweet about the topic '{{ topic }}' with a positive sentiment.

    Examples:
    - {% generate "write a tweet with a positive sentiment", "gpt-3.5-turbo" %}
    - {% generate "write a tweet with a sad sentiment", "gpt-3.5-turbo" %}
    - {% generate "write a tweet with a neutral sentiment", "gpt-3.5-turbo" %}

    \"\"\"

    async def task(task_id: int, sleep_time: int):
        logging.info(f"Task {task_id} is running.")
        await asyncio.sleep(sleep_time)
        logging.info(f"Task {task_id} done.")


    async def main():
        p = AsyncPrompt(prompt_template)
        # Schedule the prompt rendering along with two executions of 'task', one sleeping for 10 seconds
        # and one sleeping for 1 second
        results = await asyncio.gather(p.text({"topic": "AI frameworks"}), task(1, 10), task(2, 1))
        print("All tasks done, rendered prompt:")
        print(results[0])


    asyncio.run(main())
    ```
    """

    def __init__(self, text: str) -> None:
        super().__init__(text)

        if not async_enabled:
            msg = "Async is not enabled. Please set the environment variable 'BANKS_ASYNC_ENABLED=on' and try again."
            raise AsyncError(msg)

    async def text(self, data: Optional[dict] = None) -> str:
        data = self._get_context(data)
        result: str = await self._template.render_async(data)
        return result



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
from unittest import mock

from jinja2.loaders import FileSystemLoader, PackageLoader

from banks import env
from banks.loader import MultiLoader


def test_defaults():
    test_loader = MultiLoader()
    assert len(test_loader._loaders) == 2
    loader, prio = test_loader._loaders[0]
    assert type(loader) == PackageLoader
    assert prio == 10
    loader, prio = test_loader._loaders[1]
    assert type(loader) == FileSystemLoader
    assert prio == 20


def test_add_loader():
    test_loader = MultiLoader()

    # add a loader, default priority
    fs_loader = FileSystemLoader("")
    test_loader.add_loader(fs_loader)
    assert test_loader._loaders[2][0] == fs_loader
    assert test_loader._loaders[2][1] == 100

    # add another one with priority
    another_loader = FileSystemLoader("")
    test_loader.add_loader(another_loader, 50)
    assert test_loader._loaders[3][0] == another_loader
    assert test_loader._loaders[3][1] == 50


def test_get_source():
    test_loader = MultiLoader()
    # remove the default loader
    test_loader._loaders = []

    l1 = mock.MagicMock()
    test_loader.add_loader(l1, 50)
    l2 = mock.MagicMock()
    test_loader.add_loader(l2, 20)

    test_loader.get_source(env, "foo")
    l1.get_source.assert_not_called()
    l2.get_source.assert_called_with(env, "foo")



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
from banks.errors import MissingDependencyError

try:
    from simplemma.simplemma import text_lemmatizer

    simplemma_avail = True
except ImportError:
    simplemma_avail = False


def lemmatize(text: str) -> str:
    """
    Compute and return the lemmatization of the input. Language is hardcoded
    to English.

    Example:
    ```
    {{ 'The dog is running' | lemmatize }}
    'the dog be run'
    ```

    Note:
        Simplemma must be manually installed to use this filter
    """
    if not simplemma_avail:
        err_msg = "simplemma is not available, please install it with 'pip install simplemma'"
        raise MissingDependencyError(err_msg)

    words = text_lemmatizer(text, "en")
    return " ".join(words)



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: MIT
from jinja2 import Environment, select_autoescape

from .config import async_enabled
from .filters import lemmatize
from .loader import MultiLoader


def _add_extensions(env):
    """
    We lazily add extensions so that we can use the env in the extensions themselves if needed.

    For example, we use banks to manage the system prompt in `GenerateExtension`
    """
    from .extensions import GenerateExtension, HFInferenceEndpointsExtension

    env.add_extension(GenerateExtension)
    env.add_extension(HFInferenceEndpointsExtension)


# Init the Jinja env
env = Environment(
    loader=MultiLoader(),
    autoescape=select_autoescape(
        enabled_extensions=("html", "xml"),
        default_for_string=False,
    ),
    trim_blocks=True,
    lstrip_blocks=True,
    enable_async=bool(async_enabled),
)

# Setup custom filters and default extensions
env.filters["lemmatize"] = lemmatize
_add_extensions(env)


def with_env(cls):
    """
    A decorator that adds an `env` attribute to the decorated class
    """
    cls.env = env
    return cls
